// This file is a modified copy of LexBasic.c
// changes: BlitzMax-Support


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>

#include "ILexer.h"
#include "Scintilla.h"
#include "SciLexer.h"

#include "PropSetSimple.h"
#include "WordList.h"
#include "LexAccessor.h"
#include "Accessor.h"
#include "StyleContext.h"
#include "CharacterSet.h"
#include "LexerModule.h"

#ifdef SCI_NAMESPACE
using namespace Scintilla;
#endif

/* Bits:
 * 1  - whitespace
 * 2  - operator
 * 4  - identifier
 * 8  - decimal digit
 * 16 - hex digit
 * 32 - bin digit
 */
static int character_classification[128] =
{
    0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  0,  0,  1,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    1,  2,  0,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  10, 2,
    60, 60, 28, 28, 28, 28, 28, 28, 28, 28, 2,  2,  2,  2,  2,  2,
    2,  20, 20, 20, 20, 20, 20, 4,  4,  4,  4,  4,  4,  4,  4,  4,
    4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  2,  2,  2,  2,  4,
    2,  20, 20, 20, 20, 20, 20, 4,  4,  4,  4,  4,  4,  4,  4,  4,
    4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  2,  2,  2,  2,  0
};

static bool IsSpace(int c) {
	return c < 128 && (character_classification[c] & 1);
}

static bool IsOperator(int c) {
	return c < 128 && (character_classification[c] & 2);
}

static bool IsIdentifier(int c) {
	return c < 128 && (character_classification[c] & 4);
}

static bool IsDigit(int c) {
	return c < 128 && (character_classification[c] & 8);
}

static bool IsHexDigit(int c) {
	return c < 128 && (character_classification[c] & 16);
}

static bool IsBinDigit(int c) {
	return c < 128 && (character_classification[c] & 32);
}

static int LowerCase(int c)
{
	if (c >= 'A' && c <= 'Z')
		return 'a' + c - 'A';
	return c;
}





static void ColouriseBasicDoc(unsigned int startPos, int length, int initStyle,
                           WordList *keywordlists[], Accessor &styler, char comment_char) {
	bool wasfirst = true, isfirst = true; // true if first token in a line
	styler.StartAt(startPos);

	StyleContext sc(startPos, length, initStyle, styler);

	// Can't use sc.More() here else we miss the last character
	for (; ; sc.Forward()) {
		if (sc.state == SCE_B_IDENTIFIER) {
			if (!IsIdentifier(sc.ch)) {
				char s[100];
				int kstates[2] = {
					SCE_B_KEYWORD,
					SCE_B_KEYWORD2,
				};
				sc.GetCurrentLowered(s, sizeof(s));
				for (int i = 0; i < 2; i++) {
					if (keywordlists[i]->InList(s)) {
						sc.ChangeState(kstates[i]);
					}
				}
				// Types, must set them as operator else they will be
				// matched as number/constant
				if (sc.Match('.') || sc.Match('$') || sc.Match('%') ||
					sc.Match('#')) {
					sc.SetState(SCE_B_OPERATOR);
				} else {
					sc.SetState(SCE_B_DEFAULT);
				}
			}
		} else if (sc.state == SCE_B_OPERATOR) {
			if (!IsOperator(sc.ch) || sc.Match('#'))
				sc.SetState(SCE_B_DEFAULT);
		} else if (sc.state == SCE_B_LABEL) {
			if (!IsIdentifier(sc.ch))
				sc.SetState(SCE_B_DEFAULT);
		} else if (sc.state == SCE_B_CONSTANT) {
			if (!IsIdentifier(sc.ch))
				sc.SetState(SCE_B_DEFAULT);
		} else if (sc.state == SCE_B_NUMBER) {
			if (!IsDigit(sc.ch))
				sc.SetState(SCE_B_DEFAULT);
		} else if (sc.state == SCE_B_HEXNUMBER) {
			if (!IsHexDigit(sc.ch))
				sc.SetState(SCE_B_DEFAULT);
		} else if (sc.state == SCE_B_BINNUMBER) {
			if (!IsBinDigit(sc.ch))
				sc.SetState(SCE_B_DEFAULT);
		} else if (sc.state == SCE_B_STRING) {
			if (sc.ch == '"') {
				sc.ForwardSetState(SCE_B_DEFAULT);
			}
			if (sc.atLineEnd) {
				sc.SetState(SCE_B_DEFAULT);
			}
		} else if (sc.state == SCE_B_COMMENT || sc.state == SCE_B_PREPROCESSOR) {
			if (sc.atLineEnd) {
				sc.SetState(SCE_B_DEFAULT);
			}
		}

		if (sc.atLineStart)
			isfirst = true;

		if (sc.state == SCE_B_DEFAULT || sc.state == SCE_B_ERROR) {
			if (isfirst && sc.Match('#')) {
				sc.SetState(SCE_B_LABEL);
			}else if (isfirst && (sc.Match("rem ") || sc.Match("rem\n") || sc.Match("rem\r\n"))){
				sc.SetState(SCE_B_COMMENT_MULTILINE);
			} else if (sc.Match('?')) {
				sc.SetState(SCE_B_PREPROCESSOR);
			} else if (sc.Match(comment_char)) {
				sc.SetState(SCE_B_COMMENT);
			} else if (sc.Match('"')) {
				sc.SetState(SCE_B_STRING);
			} else if (IsDigit(sc.ch)) {
				sc.SetState(SCE_B_NUMBER);
			} else if (sc.Match('$')) {
				sc.SetState(SCE_B_HEXNUMBER);
			} else if (sc.Match('%')) {
				sc.SetState(SCE_B_BINNUMBER);
			} else if (IsOperator(sc.ch)) {
				sc.SetState(SCE_B_OPERATOR);
			} else if (IsIdentifier(sc.ch)) {
				wasfirst = isfirst;
				sc.SetState(SCE_B_IDENTIFIER);
			} else if (!IsSpace(sc.ch)) {
				sc.SetState(SCE_B_ERROR);
			}
		}else if (sc.state == SCE_B_COMMENT_MULTILINE){
			bool endrem = sc.Match("endrem");
			bool end_rem = sc.Match("end rem");
			if (isfirst && (endrem || end_rem)){
				while(!sc.atLineEnd){
					sc.Forward();
				}
				sc.SetState(SCE_B_DEFAULT);
			}
		}

		if (!IsSpace(sc.ch))
			isfirst = false;

		if (!sc.More())
			break;
	}
	sc.Complete();
}

static int CheckFoldPoint(char const *token, int &level, Accessor &styler, unsigned int start) {
	//I can't sleep with this dirty 'abstract'-Hack :(
	char word[256];
	int wordlen=0;
	for(int i=start;true;i++){
		int c = styler.SafeGetCharAt(i);
		if(c=='\n') break;
		word[wordlen] = static_cast<char>(LowerCase(c));
		wordlen ++;
	}
	bool abstract = (strstr(word,"abstract") != NULL);
	//forget it... look at the rest!!!
	
	if (
		(!strcmp(token, "function") && !abstract) ||
		!strcmp(token, "type")     ||
		(!strcmp(token, "method") && !abstract)   ||
		!strcmp(token, "repeat")   ||
		!strcmp(token, "while")    ||
		!strcmp(token, "for")      ||
		!strcmp(token, "try")
	    ) {
		level |= SC_FOLDLEVELHEADERFLAG;
		return 1;
	}
	if (
		!strcmp(token, "end function") ||
		!strcmp(token, "endfunction") ||
		!strcmp(token, "end type")    ||
		!strcmp(token, "endtype")     ||
		!strcmp(token, "end method")  ||
		!strcmp(token, "endmethod")   ||
		!strcmp(token, "until")       ||
		!strcmp(token, "forever")     ||
		!strcmp(token, "wend")        ||
		!strcmp(token, "endwhile")    ||
		!strcmp(token, "next")        ||
		!strcmp(token, "end try")     ||
		!strcmp(token, "endtry")
	    ) {
		return -1;
	}
	return 0;
}



static void FoldBasicDoc(unsigned int startPos, int length,
	Accessor &styler, int (*CheckFoldPoint)(char const *, int &, Accessor &, unsigned int)) {
	int line = styler.GetLine(startPos);
	int level = styler.LevelAt(line);
	int go = 0, done = 0;
	int endPos = startPos + length;
	char word[256];
	int wordlen = 0;
	int i;
        bool foldCompact = styler.GetPropertyInt("fold.compact", 1) != 0;
	// Scan for tokens at the start of the line (they may include
	// whitespace, for tokens like "End Function"
	for (i = startPos; i < endPos; i++) {
		int c = styler.SafeGetCharAt(i);
		int style = styler.StyleAt(i);
		if (!done && !go) {
			if (wordlen) { // are we scanning a token already?
				word[wordlen] = static_cast<char>(LowerCase(c));
				if (!IsIdentifier(c)) { // done with token
					word[wordlen] = '\0';
					if(style==SCE_B_COMMENT || style==SCE_B_COMMENT_MULTILINE){
						go = 0;
					}else{
						go = CheckFoldPoint(word, level, styler, i);
					}
					if (!go) {
						// Treat any whitespace as single blank, for
						// things like "End   Function".
						if (IsSpace(c) && IsIdentifier(word[wordlen - 1])) {
							word[wordlen] = ' ';
							if (wordlen < 255)
								wordlen++;
						}
						else // done with this line
							done = 1;
					}
				} else if (wordlen < 255) {
					wordlen++;
				}
			} else { // start scanning at first non-whitespace character
				if (!IsSpace(c)) {
					if (IsIdentifier(c)) {
						word[0] = static_cast<char>(LowerCase(c));
						wordlen = 1;
					} else // done with this line
						done = 1;
				}
			}
		}
		if (c == '\n') { // line end
			if (!done && wordlen == 0 && foldCompact) // line was only space
				level |= SC_FOLDLEVELWHITEFLAG;
			if (level != styler.LevelAt(line))
				styler.SetLevel(line, level);
			level += go;
			line++;
			// reset state
			wordlen = 0;
			level &= ~SC_FOLDLEVELHEADERFLAG;
			level &= ~SC_FOLDLEVELWHITEFLAG;
			go = 0;
			done = 0;
		}
	}
}



static void ColouriseDoc(unsigned int startPos, int length, int initStyle, WordList *keywordlists[], Accessor &styler) {
	ColouriseBasicDoc(startPos, length, initStyle, keywordlists, styler, '\'');
}

static void FoldDoc(unsigned int startPos, int length, int, WordList *[], Accessor &styler) {
	FoldBasicDoc(startPos, length, styler, CheckFoldPoint);
}



static const char * const WordListDesc[] = {
	"BlitzMax Runtime keywords",
	"BlitzMax module added keywords",
	0
};





LexerModule lmBlitzMax(SCLEX_BLITZMAX, ColouriseDoc,    "blitzmax", FoldDoc,    WordListDesc);




